import os
from typing import Dict
import pathlib

import sh
from cement import App, TestApp, init_defaults
from cement.core.exc import CaughtSignal
from cement.utils import fs

from tank.core.cloud_settings import CloudUserSettings
from tank.core.exc import TankError
from tank.controllers.base import Base
from tank.controllers.cluster import Cluster


def _default_config() -> Dict:
    config = init_defaults('tank',
                           'log.logging')

    config['tank'] = {
        'terraform_run_command': '/usr/local/bin/terraform',
        'terraform_inventory_run_command': '/usr/local/bin/terraform-inventory',
    }

    config['log.logging']['level'] = 'WARNING'

    return config


class MixbytesTank(App):
    """MixBytes Tank primary application."""

    class Meta:
        label = 'tank'

        # configuration defaults
        config_defaults = _default_config()

        # call sys.exit() on close
        close_on_exit = True

        # load additional framework extensions
        extensions = [
            'yaml',
            'colorlog',
            'jinja2',
        ]

        # List of configuration directory
        config_dirs = ['~/.tank']

        # configuration handler
        config_handler = 'yaml'

        # configuration file suffix
        config_file_suffix = '.yml'

        # set the log handler
        log_handler = 'colorlog'

        # set the output handler
        output_handler = 'jinja2'

        template_handler = 'jinja2'

        # register handlers
        handlers = [
            Base,
            Cluster
        ]

        # register hooks
        hooks = [
        ]


    def __init__(self):
        super().__init__()
        self._cloud_settings = None


    def setup(self):
        super(MixbytesTank, self).setup()
        fs.ensure_dir_exists(self.user_dir)

    @property
    def app_env(self) -> Dict:
        env = os.environ.copy()
        env["TF_LOG"] = "TRACE"
        env["TF_IN_AUTOMATION"] = "true"
        return env

    @property
    def cloud_settings(self) -> CloudUserSettings:
        if self._cloud_settings is None:
            self._cloud_settings = CloudUserSettings(self.config)

        return self._cloud_settings

    @property
    def terraform_run_command(self) -> str:
        return self.config.get(self.Meta.label, 'terraform_run_command')

    @property
    def terraform_inventory_run_command(self) -> str:
        return self.config.get(self.Meta.label, 'terraform_inventory_run_command')

    @property
    def user_dir(self) -> str:
        return fs.abspath(fs.join(pathlib.Path.home(), '.tank'))


    def _check_terraform_availability(self):
        try:
            sh.Command(self.terraform_run_command, "--version")
        except Exception:
            raise TankError("Error calling Terraform at '{}'".format(self.terraform_run_command))

    def _check_terraform_inventory_availability(self):
        try:
            sh.Command(self.terraform_inventory_run_command, "-version")
        except Exception:
            raise TankError("Error calling Terraform Inventory at '{}'".format(
                self.terraform_inventory_run_command))


class MixbytesTankTest(TestApp, MixbytesTank):
    """A sub-class of MixbytesTank that is better suited for testing."""

    class Meta:
        label = 'tank'


def main():
    with MixbytesTank() as app:
        try:
            app._check_terraform_availability()
            app._check_terraform_inventory_availability()
            app.run()

        except TankError as e:
            print('{}: {}'.format(e.__class__.__name__, e))
            app.exit_code = 1

            if app.debug is True:
                import traceback
                traceback.print_exc()

        # FIXME better signal handling
        except CaughtSignal as e:
            # Default Cement signals are SIGINT and SIGTERM, exit 0 (non-error)
            print('\n%s' % e)
            app.exit_code = 0


if __name__ == '__main__':
    main()



#
#   module tank.core.testcase
#

from typing import Dict

import yaml
import jsonschema

from tank.core.exc import TankTestCaseError
from tank.core.utils import yaml_load, yaml_dump


class TestCase:
    """
    Entity describing single test performed by Tank.
    """
    def __init__(self, filename):
        self.filename = filename

        self._content = yaml_load(filename)
        try:
            jsonschema.validate(self._content, self.__class__._TESTCASE_SCHEMA)
        except jsonschema.ValidationError as e:
            raise TankTestCaseError('Failed to validate testcase {}'.format(filename), e)

        for name, cfg in self._content['instances'].items():
            if name.lower() == 'monitoring':
                raise TankTestCaseError('\'monitoring\' instance name is reserved')
            if isinstance(cfg, int):
                self._content['instances'][name] = {'count': cfg, 'type': 'small'}

    def save(self, filename):
        yaml_dump(filename, self._content)

    @property
    def binding(self) -> str:
        return self._content['binding']

    @property
    def instances(self) -> Dict:
        return dict(self._content['instances'])

    @property
    def total_instances(self) -> int:
        return sum(cfg['count'] for cfg in self.instances.values())


    _TESTCASE_SCHEMA = yaml.safe_load(r'''
type: object

additionalProperties: False
required:
    - binding
    - instances

properties:

    binding:
        type: string
    
    instances:
        type: object
        propertyNames:
            pattern: "^[A-Za-z_]+$"

        additionalProperties:
        
            oneOf:
                - {type: integer, minimum: 0}
                  
                - type: object
                  additionalProperties: False
                  required:
                      - count
                  properties:
                      count: {type: integer, minimum: 0}
                      type: {enum: ["micro", "small", "standard", "large", "xlarge", "xxlarge", "huge"]}
    ''')



#
#   module tank.core.utils
#
# Misc. utils.
#

import os
import re

import yaml


def yaml_load(filename: str):
    with open(filename) as fh:
        return yaml.safe_load(fh)


def yaml_dump(filename: str, data):
    with open(filename, 'w') as fh:
        return yaml.dump(data, fh, default_flow_style=False)


def grep_dir(dirname: str, filter_regex: str = None, isdir: bool = False):
    """
    Enumerate and filter contents of a directory.
    """
    contents = os.listdir(dirname)

    if filter_regex is not None:
        filter_re = re.compile(filter_regex)
        contents = filter(lambda name: filter_re.match(name) is not None, contents)

    if isdir is not None:
        contents = filter(lambda name: os.path.isdir(os.path.join(dirname, name)), contents)

    return contents




from cement import Controller, ex
from tabulate import tabulate

from tank.core.run import Run
from tank.core.testcase import TestCase
from tank.core.lambdas import first


class Cluster(Controller):

    class Meta:
        label = 'cluster'
        stacked_type = 'nested'
        stacked_on = 'base'

        # text displayed at the top of --help output
        description = 'Manipulating a cluster'

        # text displayed at the bottom of --help output
        title = 'Low level cluster management commands'
        help = 'Low level cluster management commands'

    @ex(help='Show clusters')
    def list(self):
        runs = Run.list_runs(self.app)

        def make_row(run):
            return [
                run.run_id,
                run.created_at.strftime('%c'),
                run.testcase_copy.total_instances + 1,
                run.meta['testcase_filename']
            ]

        print(tabulate(list(map(make_row, runs)), headers=['RUN ID', 'CREATED', 'INSTANCES', 'TESTCASE']))

    @ex(help='Init a Tank run, download plugins and modules for Terraform', hide=True,
        arguments=[(['testcase'], {'type': str, 'nargs': 1})])
    def init(self):
        testcase = TestCase(first(self.app.pargs.testcase))
        run = Run.new_run(self.app, testcase)
        print('Created tank run: {}'.format(run.run_id))

        run.init()

    @ex(help='Generate and show an execution plan by Terraform', hide=True,
        arguments=[(['run_id'], {'type': str, 'nargs': 1})])
    def plan(self):
        Run(self.app, first(self.app.pargs.run_id)).plan()

    @ex(help='Create instances for cluster', hide=True,
        arguments=[(['run_id'], {'type': str, 'nargs': 1})])
    def create(self):
        Run(self.app, first(self.app.pargs.run_id)).create()

    @ex(help='Install Ansible roles from Galaxy or SCM', hide=True,
        arguments=[(['run_id'], {'type': str, 'nargs': 1})])
    def dependency(self):
        Run(self.app, first(self.app.pargs.run_id)).dependency()

    @ex(help='Setup instances: configs, packages, services, etc', hide=True,
        arguments=[(['run_id'], {'type': str, 'nargs': 1})])
    def provision(self):
        Run(self.app, first(self.app.pargs.run_id)).provision()

    @ex(help='Runs bench on prepared cluster',
        arguments=[
            (['run_id'],
             {'type': str, 'nargs': 1}),
            (['--tps'],
             {'help': 'set global transactions per second generation rate',
              'type': int}),
            (['--total-tx'],
             {'help': 'how many transactions to send',
              'type': int}),
        ])
    def bench(self):
        Run(self.app, first(self.app.pargs.run_id)).bench(self.app.pargs.tps, self.app.pargs.total_tx)

    @ex(help='Destroy all instances of the cluster',
        arguments=[(['run_id'], {'type': str, 'nargs': 1})])
    def destroy(self):
        Run(self.app, first(self.app.pargs.run_id)).destroy()

    @ex(help='Create and setup a cluster (init, create, dependency, provision)',
        arguments=[(['testcase'], {'type': str, 'nargs': 1})])
    def deploy(self):
        testcase = TestCase(first(self.app.pargs.testcase))
        run = Run.new_run(self.app, testcase)
        print('Created tank run: {}'.format(run.run_id))

        run.init()
        run.create()
        run.dependency()
        run.provision()
